/*
   Dot3 Specular Bump Mapping using Hlsl.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is build off of the first Hlsl Dot3 bump mapping
   demo and add the ability to have specular highlights.  To keep things simple
   all the necessary information is hardcoded in the effect shader file.
*/


// Demo: Dot3 Specular Bump Mapping using Hlsl.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);
bool InitializeObject();
void RenderScene();
void ShutdownDirect3D();


// DirectX objects.
LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;

// Textures.
LPDIRECT3DTEXTURE9 Texture = NULL;
LPDIRECT3DTEXTURE9 bumpMap = NULL;

// Shader and vertex decalaration.
LPD3DXEFFECT effectShader = NULL;
LPDIRECT3DVERTEXDECLARATION9 ShaderDeclaration = NULL;

// Shader declaration.
D3DVERTEXELEMENT9 SDec[] =
{
   {0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0},
   {0, 12, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL, 0},
	{0, 24, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0},
   {0, 32, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT, 0},
   {0, 44, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL, 0},
   D3DDECL_END()
};

// Our model object.
LPD3DXMESH Model = NULL;

// Direction of our main light.
D3DXVECTOR4 lightPos = D3DXVECTOR4(0.0f, 0.0f, -10.0f, 0.0f);

// Our main 3 matrices...
D3DXMATRIX WorldMatrix;
D3DXMATRIX ViewMatrix;
D3DXMATRIX ProjectionMatrix;

// Angle is used to rotate the object.
// gWidth and gHeight is used for the aspect ratio.
float angle = 0.0;
int gWidth = 0;
int gHeight = 0;


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   int width, height;

   switch(message)
      {
         case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;

         case WM_CLOSE:
         case WM_DESTROY:
            PostQuitMessage(0);
            break;

         case WM_SIZE:
            height = HIWORD(lParam);
            width = LOWORD(lParam);

            if(height == 0) height = 1;
            if(width == 0) width = 1;

            gWidth = width;
            gHeight = height;
            break;

         default:
         break;
      }

    return DefWindowProc(hwnd, message, wParam, lParam);
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,                              // The extended style.
                         "UGPClass",                        // Window class.
                         "Hlsl Specular Dot3 Bump - by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |// Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |     // Window style.
                          WS_CLIPSIBLINGS,                  // Window style.
                          100, 100,                         // X, Y coords.
                          640, 480,                         // Window size.
                          NULL,                             // Handle to parent window.
                          NULL,                             // Menu.
                          hInstance,                        // Handle to app instance.
                          NULL);                            // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT,
	                                         D3DDEVTYPE_HAL, &D3DCaps)))
		{
		   return false;
	   }

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = 640;
         Present_Parameters.BackBufferHeight = 480;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode.
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;
   Present_Parameters.BackBufferFormat = DisplayMode.Format;
   Present_Parameters.BackBufferCount = 1;
   Present_Parameters.EnableAutoDepthStencil = TRUE;
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                                           D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                           &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }

   // One last check to be sure.
   if(D3D_Device == NULL)
      {
         MessageBox(NULL, "D3D_Device is equal to NULL!?!", "Error!", MB_OK);
         return false;
      }

   // Set our states.
   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);
   D3D_Device->SetRenderState(D3DRS_DITHERENABLE, TRUE);
   D3D_Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	D3D_Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

   // We can use this to store any errors when compiling the shader effect file.
	LPD3DXBUFFER errors = NULL;

   // Load the effect file.
	if(FAILED(D3DXCreateEffectFromFile(D3D_Device, "effect.fx", NULL, NULL, 0,
                                      NULL, &effectShader, &errors)))
      {
		   LPVOID compileErrors = errors->GetBufferPointer();
         MessageBox(NULL, (const char*)compileErrors, "Shader Compile Error...",
                    MB_OK | MB_ICONEXCLAMATION);
         return false;
	   }

   // Initialize out .x model.
   if(!InitializeObject()) return false;

   // Set the projection matrix.
   D3DXMatrixPerspectiveFovLH(&ProjectionMatrix, 45.0f,
                              (float)gWidth/(float)gHeight, 0.1f, 1000.0f);
   D3D_Device->SetTransform(D3DTS_PROJECTION, &ProjectionMatrix);

   return true;
}


bool InitializeObject()
{
   // Load the model from the .x file we will bump map.
   if(FAILED(D3DXLoadMeshFromX("sphere.x", D3DXMESH_SYSTEMMEM,
                               D3D_Device, NULL, NULL, NULL, NULL, &Model)))
      {
         MessageBox(NULL, "Error loading Model.x model file...", "Error!", MB_OK);
         return false;
      }

   // Copy declaration into mesh model.
   Model->CloneMesh(D3DXMESH_MANAGED, SDec, D3D_Device, &Model);

   // Calculate all the normals of the model.
   if(FAILED(D3DXComputeNormals(Model, NULL))) return false;

   // Calculate the tangents and binormals.
   if(FAILED(D3DXComputeTangent(Model, 0, 0, 0, true, NULL))) return false;

   // Load the decal image.
   if(D3DXCreateTextureFromFile(D3D_Device, "decal.tga", &Texture) != D3D_OK)
      return false;

   // Load bump map.
   if(D3DXCreateTextureFromFile(D3D_Device, "normal.tga", &bumpMap) != D3D_OK)
      return false;

   return true;
}


void SetShader()
{
   // This will create a rotation matrix.
   D3DXMatrixRotationY(&WorldMatrix, angle);

   // Add 0.003 to angle to rotate the object even more the next frame.
   angle += 0.003f;
   if(angle >= 360) angle = 0.0f;

   // Create a view matrix (set the camera).
   D3DXVECTOR3 pos(0.0f, 0.0f, -2.5f);
   D3DXVECTOR3 lookat(0.0f, 0.0f, 0.0f);
   D3DXVECTOR3 up(0.0f, 1.0f, 0.0f);
   D3DXMatrixLookAtLH(&ViewMatrix, &pos, &lookat, &up);

   // Set the model view projection matrix.
   D3DXMATRIX modelViewProj = WorldMatrix * ViewMatrix * ProjectionMatrix;

   // Set the shader parameters.
   D3DXMatrixTranspose(&modelViewProj, &modelViewProj);
   effectShader->SetMatrix("worldViewProj", &modelViewProj);
   
   // Set the world matrix.
   effectShader->SetMatrix("worldView", &WorldMatrix);

   // Set the light position.
   effectShader->SetVector("lightPos", &lightPos);

   // Set how the vertex shader is defined.
   D3D_Device->SetVertexDeclaration(ShaderDeclaration);
}


void RenderScene()
{
   // Clear the screen.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   // Tell Direct3D we are now going to start drawing.
   D3D_Device->BeginScene();

      // Apply the textures.
      effectShader->SetTechnique("SpecularBumpMapping");
      effectShader->SetTexture("decal", Texture);
      effectShader->SetTexture("bump", bumpMap);

      // Set the vertex shader variables.
      SetShader();
      
      effectShader->Begin(NULL, 0);

         effectShader->BeginPass(0);

            // Draw the model.
            Model->DrawSubset(0);

         effectShader->EndPass();

      // End of the effect.
      effectShader->End();

   // We are done drawing for this scene.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects and vertex shader.
   if(Texture != NULL) 
      {
         Texture->Release();
         Texture = NULL;
      }
   
   if(bumpMap != NULL) 
      {
         bumpMap->Release();
         bumpMap = NULL;
      }

   if(Model != NULL)
      {
         Model->Release();
         Model = NULL;
      }

   if(effectShader != NULL)
      {
         effectShader->Release();
         effectShader = NULL;
      }

   if(ShaderDeclaration != NULL)
      {
         ShaderDeclaration->Release();
         ShaderDeclaration = NULL;
      }

   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }
}


// Recap:

// With a few extra lines we were able to add specular highlights to give our
// object a shinny effect.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com